# Copyright 2007 Collabora Ltd.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

import dbus
import hippo
import gtk
import pango

from sugar import profile
from sugar.activity.activity import Activity
from sugar.graphics.canvasicon import CanvasIcon
from sugar.graphics import color
from sugar.graphics.xocolor import XoColor
from sugar.presence import presenceservice

from telepathy.client import Connection, Channel

from telepathy.interfaces import (
    CONN_INTERFACE, PROPERTIES_INTERFACE, CONN_INTERFACE_PRESENCE,
    CHANNEL_INTERFACE_GROUP, CONN_INTERFACE_ALIASING,
    CHANNEL_TYPE_TEXT, CHANNEL_TYPE_CONTACT_LIST, CHANNEL_INTERFACE,
    CHANNEL_TYPE_STREAMED_MEDIA, CONN_INTERFACE_CAPABILITIES)

CONN_INTERFACE_BUDDY_INFO = 'org.laptop.Telepathy.BuddyInfo'

from telepathy.constants import (
    CONNECTION_HANDLE_TYPE_NONE, CONNECTION_HANDLE_TYPE_CONTACT,
    CONNECTION_HANDLE_TYPE_ROOM, CHANNEL_TEXT_MESSAGE_TYPE_NORMAL,
    CONNECTION_STATUS_CONNECTED, CONNECTION_STATUS_DISCONNECTED,
    CONNECTION_STATUS_CONNECTING, CONNECTION_HANDLE_TYPE_LIST,
    MEDIA_STREAM_TYPE_AUDIO, MEDIA_STREAM_TYPE_VIDEO,
    MEDIA_STREAM_STATE_DISCONNECTED, MEDIA_STREAM_STATE_CONNECTING,
    MEDIA_STREAM_STATE_CONNECTED,
    CONNECTION_CAPABILITY_FLAG_CREATE, CONNECTION_CAPABILITY_FLAG_INVITE)

def get_stream_engine():
    bus = dbus.Bus()
    return bus.get_object(
        'org.freedesktop.Telepathy.StreamEngine',
        '/org/freedesktop/Telepathy/StreamEngine')


class Call(Activity):
    def __init__(self, handle):
        Activity.__init__(self, handle)

        self.set_title('Call')

        root = self.make_root()
        self.set_canvas(root)
        root.show_all()
        self.contacts = {}
        self.incomings = {}

        self.owner_color = profile.get_color()
        self.owner_nickname = profile.get_nick_name()

        pservice = presenceservice.get_instance()

        bus = dbus.Bus()
        name, path = pservice.get_preferred_connection()
        conn = Connection(name, path)

        # hack
        conn._valid_interfaces.add(CONN_INTERFACE_PRESENCE)
        conn._valid_interfaces.add(CONN_INTERFACE_ALIASING)
        conn._valid_interfaces.add(CONN_INTERFACE_BUDDY_INFO)
        conn._valid_interfaces.add(CONN_INTERFACE_CAPABILITIES)

        conn[CONN_INTERFACE].connect_to_signal('StatusChanged',
                self.status_changed_cb)
        conn[CONN_INTERFACE].connect_to_signal('NewChannel',
                self.new_channel_cb)
        conn[CONN_INTERFACE_PRESENCE].connect_to_signal('PresenceUpdate',
                self.presence_update_cb)
        conn[CONN_INTERFACE_CAPABILITIES].connect_to_signal(
                'CapabilitiesChanged',
                self.cap_changed_cb)

        conn[CONN_INTERFACE_CAPABILITIES].AdvertiseCapabilities(
                [(CHANNEL_TYPE_STREAMED_MEDIA, 3)], [])

        self.conn = conn
        self.contact = 0
        self.channel = None
        self.caller = False

        status = conn[CONN_INTERFACE].GetStatus()
        if status == CONNECTION_STATUS_CONNECTED:
            print "connected"
            self.create_list ()

        self.connect("destroy", self.destroy)

        se = dbus.Interface(get_stream_engine(),
            'org.freedesktop.Telepathy.StreamEngine')

        se.AddPreviewWindow(self.preview.get_id())

    def destroy (self, *args):
        # activity closed, we can't deal with media streams anymore
        self.conn[CONN_INTERFACE_CAPABILITIES].AdvertiseCapabilities(
                [], [CHANNEL_TYPE_STREAMED_MEDIA])

    def check_caps (self, caps):
        for contact, channel_type, generic_flags, specific_flags in caps:
            if channel_type == CHANNEL_TYPE_STREAMED_MEDIA and\
                    generic_flags == (CONNECTION_CAPABILITY_FLAG_CREATE |
                            CONNECTION_CAPABILITY_FLAG_INVITE):
                return True
        return False

    def presence_update_cb (self, presence):
        for handle in presence:
            timestamp, statuses = presence[handle]
            for status, params in statuses.items():
                if status in ["available", "away", "brb", "busy", "dnd", "xa"]:
                    caps = self.conn[CONN_INTERFACE_CAPABILITIES].GetCapabilities([handle])
                    if self.check_caps (caps):
                        self.add_contact (handle)
                else:
                    self.remove_contact (handle)

    def cap_changed_cb (self, caps):
        for contact, type, old_generic_flags, new_generic_flags,\
                old_specific_flags, new_specific_flags in caps:
            if type == CHANNEL_TYPE_STREAMED_MEDIA:
                if new_generic_flags == (CONNECTION_CAPABILITY_FLAG_CREATE |
                        CONNECTION_CAPABILITY_FLAG_INVITE):
                    self.add_contact (contact)
                else:
                    self.remove_contact (contact)
                break

    def create_list(self):
        handle = self.conn[CONN_INTERFACE].RequestHandles(
            CONNECTION_HANDLE_TYPE_LIST, ['subscribe'])[0]
        chan_path = self.conn[CONN_INTERFACE].RequestChannel(
            CHANNEL_TYPE_CONTACT_LIST, CONNECTION_HANDLE_TYPE_LIST,
            handle, True)
        channel = Channel(self.conn._dbus_object._named_service, chan_path)

        handles = channel[CHANNEL_INTERFACE_GROUP].GetMembers()

        self.conn[CONN_INTERFACE_PRESENCE].RequestPresence (handles)

    def status_changed_cb(self, status, reason):
        if status == CONNECTION_STATUS_CONNECTED:
            pass
        elif status == CONNECTION_STATUS_DISCONNECTED:
            print 'disconnected'
            if self.channel is not None:
                self.channel.Close()

    def make_root(self):
        output_frame = gtk.Frame()
        output_frame.set_shadow_type(gtk.SHADOW_IN)
        self.output = gtk.Socket()
        self.output.set_size_request(400, 300)
        output_frame.add(self.output)

        # XXX : this is crack to display the output widget centered on the
        # screen and not extended.
        # There should be a better way to do that.
        table = gtk.Table(2, 1, homogeneous=False)
        table.attach(gtk.Frame(), 0, 1, 0, 1)
        table.attach(output_frame, 0, 2, 0, 1, xoptions=0, yoptions=0)

        preview_frame = gtk.Frame()
        preview_frame.set_shadow_type(gtk.SHADOW_IN)
        self.preview = gtk.Socket()
        self.preview.set_size_request(200, 150)
        preview_frame.add(self.preview)

        self.incomings_box = hippo.CanvasBox(spacing=4,
                orientation=hippo.ORIENTATION_HORIZONTAL)

        self.contact_box = hippo.CanvasBox(spacing=4,
                orientation=hippo.ORIENTATION_HORIZONTAL)
        widget = hippo.CanvasWidget(widget=preview_frame)
        self.contact_box.append(widget)

        box = hippo.CanvasBox()
        widget = hippo.CanvasWidget(widget=table)
        box.append(widget, hippo.PACK_EXPAND)
        box.append(self.incomings_box)
        box.append(self.contact_box)

        canvas = hippo.Canvas()
        canvas.set_root(box)

        return canvas

    def make_owner_icon(self):
        return CanvasIcon(
            icon_name='theme:stock-buddy',
            xo_color=self.owner_color)

    def request_alias_reply_cb(self, handle, aliases):
        try:
            icon, name, vbox = self.contacts[handle]
            name.props.text = aliases[0]
        except KeyError:
            pass

    def request_alias_error_cb(self, handle, *args):
        pass

    def get_properties_reply_cb(self, handle, infos):
        try:
            icon, name, vbox = self.contacts[handle]
            icon.props.xo_color = XoColor(infos['color'])
        except KeyError:
            pass

    def get_properties_error_cb(self, handle, *args):
        pass

    def add_contact(self, handle):
        if handle == self.conn[CONN_INTERFACE].GetSelfHandle():
            return

        if self.contacts.get(handle):
            return self.contacts.get(handle)

        jid = self.conn[CONN_INTERFACE].InspectHandles(
                CONNECTION_HANDLE_TYPE_CONTACT, [handle])[0]
        alias = jid.split('@')[0]
        self.conn[CONN_INTERFACE_ALIASING].RequestAliases([handle],
            reply_handler=lambda aliases: self.request_alias_reply_cb(handle, aliases),
            error_handler=lambda *args: self.request_alias_error_cb(handle, *args))

        print "add %d %s (%s)" % (handle, jid, alias)

        self.conn[CONN_INTERFACE_BUDDY_INFO].GetProperties(handle,
            reply_handler=lambda infos: self.get_properties_reply_cb(handle, infos),
            error_handler=lambda *args: self.get_properties_error_cb(handle, *args))

        buddy_color = "#000000,#ffffff"
        icon = CanvasIcon(
            icon_name='theme:stock-buddy',
            xo_color=XoColor(buddy_color))

        icon.connect('activated', self.icon_activated_cb, handle)

        name = hippo.CanvasText(text=alias, color=color.WHITE.get_int())
        vbox = hippo.CanvasBox(padding=5)

        vbox.append(icon)
        vbox.append(name)

        self.contact_box.append(vbox)
        self.contacts[handle] = icon, name, vbox

        return self.contacts[handle]

    def remove_contact (self, handle):
        widgets = self.contacts.get(handle)
        if widgets:
            icon, name, vbox = widgets
            print "remove %d" % handle
            self.contact_box.remove(vbox)
            del self.contacts[handle]

        if self.channel is not None and self.contact == handle:
            print "contact leave. Close the call"
            self.channel.Close()

        incoming = self.incomings.get(handle)
        if incoming is not None:
            self.incomings_box.remove(self.incomings[handle])
            del self.incomings[handle]

    def icon_activated_cb (self, item, handle):
        # call the contact
        if self.contact == handle:
            print "we are already calling %d" % handle
            return

        if self.channel is not None:
            self.channel.Close()

        self.caller = True
        self.contact = handle
        print "call contact %d" % handle
        self.conn[CONN_INTERFACE].RequestChannel(
                 CHANNEL_TYPE_STREAMED_MEDIA, CONNECTION_HANDLE_TYPE_NONE,
                 0, True)

    def new_channel_cb(self, object_path, channel_type, handle_type, handle,
            suppress_handler):
        if channel_type != CHANNEL_TYPE_STREAMED_MEDIA:
            return

        self.chan_handle_type = handle_type
        self.chan_handle = handle

        print "new streamed media channel"
        Channel(self.conn._dbus_object._named_service, object_path,
                ready_handler=self.channel_ready_cb)

    def channel_ready_cb(self, channel):
        print "channel ready"
        channel[CHANNEL_INTERFACE].connect_to_signal('Closed', self.closed_cb)
        channel[CHANNEL_INTERFACE_GROUP].connect_to_signal('MembersChanged',
            self.members_changed_cb)
        channel[CHANNEL_TYPE_STREAMED_MEDIA].connect_to_signal(
            'StreamAdded', self.stream_added_cb)
        channel[CHANNEL_TYPE_STREAMED_MEDIA].connect_to_signal(
            'StreamStateChanged', self.stream_state_changed_cb)

        stream_engine = get_stream_engine()
        handler = dbus.Interface(stream_engine,
            'org.freedesktop.Telepathy.ChannelHandler')
        handler.HandleChannel(
            self.conn._dbus_object._named_service,
            self.conn.object_path,
            CHANNEL_TYPE_STREAMED_MEDIA,
            channel.object_path,
            self.chan_handle_type,
            self.chan_handle)

        self.channel = channel

        if self.caller:
            # we initiate the call
            channel[CHANNEL_INTERFACE_GROUP].AddMembers([self.contact], "")

            print "requesting audio/video streams"

            try:
                channel[CHANNEL_TYPE_STREAMED_MEDIA].RequestStreams(
                    self.contact,
                    [MEDIA_STREAM_TYPE_AUDIO, MEDIA_STREAM_TYPE_VIDEO]);
            except dbus.DBusException, e:
                print "failed:", e
                print "requesting audio stream"

                try:
                    channel[CHANNEL_TYPE_STREAMED_MEDIA].RequestStreams(
                        self.contact, [MEDIA_STREAM_TYPE_AUDIO]);
                except dbus.DBusException, e:
                    print "failed:", e
                    print "giving up"
                    self.channel.Close()

        else:
            # we receive an incoming call
            members = channel[CHANNEL_INTERFACE_GROUP].GetMembers()
            self.incoming_call_request(members[0], channel)

            streams = channel[CHANNEL_TYPE_STREAMED_MEDIA].ListStreams()
            for id, contact_handle, type, state, direction, flag in streams:
                if type == MEDIA_STREAM_TYPE_AUDIO:
                    print "audio stream"
                    self.audio_stream = id

                if type == MEDIA_STREAM_TYPE_VIDEO:
                    print "video stream"
                    self.video_stream = id

    def closed_cb(self):
        print "channel with %d closed" % self.contact
        self.channel = None
        self.caller = False
        self.contact = 0

    def members_changed_cb(self, message, added, removed, local_pending,
            remote_pending, actor, reason):
        print 'MembersChanged', (
            added, removed, local_pending, remote_pending, actor, reason)

    def stream_added_cb (self, stream_id, contact_handle, stream_type):
        if contact_handle != self.contact:
            return

        if stream_type == MEDIA_STREAM_TYPE_AUDIO:
            print "add audio stream"
            self.audio_stream = stream_id

        elif stream_type == MEDIA_STREAM_TYPE_VIDEO:
            print "add video stream"
            self.video_stream = stream_id

    def stream_state_changed_cb (self, stream_id, stream_state):
        if stream_state == MEDIA_STREAM_STATE_CONNECTING and\
                stream_id == self.video_stream:
            print "Setting video output window"
            se = dbus.Interface(get_stream_engine(),
                'org.freedesktop.Telepathy.StreamEngine')

            se.SetOutputWindow(self.channel.object_path, stream_id,
                    self.output.get_id())

    def incoming_call_request(self, handle, channel):
        print "call from %d" % handle
        contact_widgets = self.contacts.get(handle)
        if contact_widgets is None:
            contact_widgets = self.add_contact (handle)

        incoming_box = hippo.CanvasBox(spacing=2,
                orientation=hippo.ORIENTATION_VERTICAL)

        incoming_box.append(hippo.CanvasText(text="Incoming Call",
            color=color.WHITE.get_int()))

        icon, name, vbox = contact_widgets
        # We have to copy the canvas text because if not, the widget is
        # removed from the contact_box when we remove the incoming_box.
        name_copy = hippo.CanvasText(text=name.props.text, color=color.WHITE.get_int())
        contact_box = hippo.CanvasBox(padding=5)
        contact_box.append(icon)
        contact_box.append(name_copy)

        incoming_box.append(contact_box)

        buttons_box = hippo.CanvasBox(spacing=4,
                orientation=hippo.ORIENTATION_HORIZONTAL)

        accept = CanvasIcon(
            icon_name='theme:stock-accept')
        accept.connect('activated', self.incoming_call_accept, handle, channel)
        buttons_box.append(accept)

        decline = CanvasIcon(
            icon_name='theme:stock-remove')
        decline.connect('activated', self.incoming_call_decline, handle, channel)
        buttons_box.append(decline)

        incoming_box.append(buttons_box)

        self.incomings_box.append(incoming_box)
        self.incomings[handle] = incoming_box

    def incoming_call_accept(self, item, handle, channel):
        print "accept incoming call"

        self.contact = handle
        pending = channel[CHANNEL_INTERFACE_GROUP].GetLocalPendingMembers()
        channel[CHANNEL_INTERFACE_GROUP].AddMembers(pending, "")

        self.incomings_box.remove(self.incomings[handle])
        del self.incomings[handle]

    def incoming_call_decline(self, item, handle, channel):
        print "decline incoming call from %d" % handle
        channel.Close()

        self.incomings_box.remove(self.incomings[handle])
        del self.incomings[handle]
